@@ -62,89 +62,6 @@ let authenticity s = Rx.match_string authenticityRx s
    SSH password interaction.
 *)
 
-(*
-let a1 = [|'p';'q';'r';'s';'t';'u';'v';'w';'x';'y';'z';'P';'Q';'R';'S';'T'|]
-let a2 = [|'0';'1';'2';'3';'4';'5';'6';'7';'8';'9';'a';'b';'c';'d';'e';'f'|]
-exception Break of (Unix.file_descr * string) option
-let ptyMasterOpen () =
-  if not(Osx.isMacOSX or Osx.isLinux) then None else
-  try
-    (* Adapted from Stevens' Advanced Programming in Unix *)
-    let x = "/dev/pty--" in
-    for i = 0 to Array.length a1 do
-      x.[8] <- a1.(i);
-      for j = 0 to Array.length a2 do
-        x.[9] <- a2.(j);
-        let fdOpt =
-          try Some(Unix.openfile x [Unix.O_RDWR] 0)
-          with Unix.Unix_error _ -> None in
-        match fdOpt with None -> ()
-        | Some fdMaster ->
-          x.[5] <- 't';
-          raise (Break(Some(fdMaster,x)))
-      done
-    done;
-    None
-  with Break z -> z
-
-let ptySlaveOpen = function
-    None -> None
-  | Some(fdMaster,ttySlave) ->
-      let slave =
-        try Some (Unix.openfile ttySlave [Unix.O_RDWR] 0o600)
-        with Unix.Unix_error _ -> None in
-      (try Unix.close fdMaster with Unix.Unix_error(_,_,_) -> ());
-      slave
-
-let printTermAttrs fd = (* for debugging *)
-  let tio = Unix.tcgetattr fd in
-  let boolPrint name x d =
-    if x then Printf.printf "%s is ON (%s)\n" name d
-    else Printf.printf "%s is OFF (%s)\n" name d in
-  let intPrint name x d =
-    Printf.printf "%s = %d (%s)\n" name x d in
-  let charPrint name x d =
-    Printf.printf "%s = '%c' (%s)\n" name x d in
-  boolPrint "c_ignbrk" tio.Unix.c_ignbrk   "Ignore the break condition.";
-  boolPrint "c_brkint" tio.Unix.c_brkint   "Signal interrupt on break condition.";
-  boolPrint "c_ignpar" tio.Unix.c_ignpar   "Ignore characters with parity errors.";
-  boolPrint "c_parmrk" tio.Unix.c_parmrk   "Mark parity errors.";
-  boolPrint "c_inpck" tio.Unix.c_inpck     "Enable parity check on input.";
-  boolPrint "c_istrip" tio.Unix.c_istrip   "Strip 8th bit on input characters.";
-  boolPrint "c_inlcr" tio.Unix.c_inlcr     "Map NL to CR on input.";
-  boolPrint "c_igncr" tio.Unix.c_igncr     "Ignore CR on input.";
-  boolPrint "c_icrnl" tio.Unix.c_icrnl     "Map CR to NL on input.";
-  boolPrint "c_ixon" tio.Unix.c_ixon       "Recognize XON/XOFF characters on input.";
-  boolPrint "c_ixoff" tio.Unix.c_ixoff     "Emit XON/XOFF chars to control input flow.";
-  boolPrint "c_opost" tio.Unix.c_opost     "Enable output processing.";
-  intPrint "c_obaud" tio.Unix.c_obaud      "Output baud rate (0 means close connection).";
-  intPrint "c_ibaud" tio.Unix.c_ibaud      "Input baud rate.";
-  intPrint "c_csize" tio.Unix.c_csize      "Number of bits per character (5-8).";
-  intPrint "c_cstopb" tio.Unix.c_cstopb    "Number of stop bits (1-2).";
-  boolPrint "c_cread" tio.Unix.c_cread     "Reception is enabled.";
-  boolPrint "c_parenb" tio.Unix.c_parenb   "Enable parity generation and detection.";
-  boolPrint "c_parodd" tio.Unix.c_parodd   "Specify odd parity instead of even.";
-  boolPrint "c_hupcl" tio.Unix.c_hupcl     "Hang up on last close.";
-  boolPrint "c_clocal" tio.Unix.c_clocal   "Ignore modem status lines.";
-  boolPrint "c_isig" tio.Unix.c_isig       "Generate signal on INTR, QUIT, SUSP.";
-  boolPrint "c_icanon" tio.Unix.c_icanon   "Enable canonical processing (line buffering and editing)";
-  boolPrint "c_noflsh" tio.Unix.c_noflsh   "Disable flush after INTR, QUIT, SUSP.";
-  boolPrint "c_echo" tio.Unix.c_echo       "Echo input characters.";
-  boolPrint "c_echoe" tio.Unix.c_echoe     "Echo ERASE (to erase previous character).";
-  boolPrint "c_echok" tio.Unix.c_echok     "Echo KILL (to erase the current line).";
-  boolPrint "c_echonl" tio.Unix.c_echonl   "Echo NL even if c_echo is not set.";
-  charPrint "c_vintr" tio.Unix.c_vintr     "Interrupt character (usually ctrl-C).";
-  charPrint "c_vquit" tio.Unix.c_vquit     "Quit character (usually ctrl-\\).";
-  charPrint "c_verase" tio.Unix.c_verase   "Erase character (usually DEL or ctrl-H).";
-  charPrint "c_vkill" tio.Unix.c_vkill     "Kill line character (usually ctrl-U).";
-  charPrint "c_veof" tio.Unix.c_veof       "End-of-file character (usually ctrl-D).";
-  charPrint "c_veol" tio.Unix.c_veol       "Alternate end-of-line char. (usually none).";
-  intPrint "c_vmin" tio.Unix.c_vmin        "Minimum number of characters to read before the read request is satisfied.";
-  intPrint "c_vtime" tio.Unix.c_vtime      "Maximum read wait (in 0.1s units).";
-  charPrint "c_vstart" tio.Unix.c_vstart   "Start character (usually ctrl-Q).";
-  charPrint "c_vstop" tio.Unix.c_vstop      "Stop character (usually ctrl-S)."
-*)
-
 (* Implemented in file pty.c *)
 type pty
 external win_openpty : unit -> (Unix.file_descr * Unix.file_descr)
@@ -160,28 +77,38 @@ external c_openpty : unit -> Unix.file_descr * Unix.file_descr =
 let openpty() = try Some (c_openpty ()) with Unix.Unix_error _ -> None
 
 (* Utility functions copied from ocaml's unix.ml because they are not exported :-| *)
-let rec safe_dup fd =
-  let new_fd = Unix.dup fd in
-  if dumpFd new_fd >= 3 then
-    new_fd
-  else begin
-    let res = safe_dup fd in
-    Unix.close new_fd;
-    res
-  end
+(* Duplicate [fd] if needed to make sure it isn't one of the
+   standard descriptors (stdin, stdout, stderr).
+   Note that this function always leaves the standard descriptors open,
+   the caller must take care of closing them if needed.
+   The "cloexec" mode doesn't matter, because
+   the descriptor returned by [dup] will be closed before the [exec],
+   and because no other thread is running concurrently
+   (we are in the child process of a fork).
+ *)
+let rec file_descr_not_standard fd =
+  if dumpFd fd >= 3 then fd else file_descr_not_standard (Unix.dup fd)
+
 let safe_close fd = try Unix.close fd with Unix.Unix_error _ -> ()
+
 let perform_redirections new_stdin new_stdout new_stderr =
-  let newnewstdin = safe_dup new_stdin in
-  let newnewstdout = safe_dup new_stdout in
-  let newnewstderr = safe_dup new_stderr in
+  let new_stdin = file_descr_not_standard new_stdin in
+  let new_stdout = file_descr_not_standard new_stdout in
+  let new_stderr = file_descr_not_standard new_stderr in
+  (*  The three dup2 close the original stdin, stdout, stderr,
+      which are the descriptors possibly left open
+      by file_descr_not_standard *)
+  Unix.dup2 ~cloexec:false new_stdin Unix.stdin;
+  Unix.dup2 ~cloexec:false new_stdout Unix.stdout;
+  Unix.dup2 ~cloexec:false new_stderr Unix.stderr;
   safe_close new_stdin;
   safe_close new_stdout;
-  safe_close new_stderr;
-  Unix.dup2 newnewstdin Unix.stdin; Unix.close newnewstdin;
-  Unix.dup2 newnewstdout Unix.stdout; Unix.close newnewstdout;
-  Unix.dup2 newnewstderr Unix.stderr; Unix.close newnewstderr
+  safe_close new_stderr
 
 let rec safe_waitpid pid =
+  (* This function is intentionally synchronous so that it can be run during
+     cleanup code when Lwt threads might be stopped or otherwise be in an
+     unreliable state. *)
   let kill_noerr si = try Unix.kill pid si with Unix.Unix_error _ -> () in
   let t = Unix.gettimeofday () in
   let rec aux st =
@@ -292,14 +219,19 @@ let unix_create_session cmd args new_stdin new_stdout new_stderr =
   match openpty () with
     None -> fallback_session cmd args new_stdin new_stdout new_stderr
   | Some (masterFd, slaveFd) ->
-(*
-      Printf.printf "openpty returns %d--%d\n" (dumpFd fdM) (dumpFd fdS); flush stdout;
-      Printf.printf "new_stdin=%d, new_stdout=%d, new_stderr=%d\n"
-        (dumpFd new_stdin) (dumpFd new_stdout) (dumpFd new_stderr) ; flush stdout;
-*)
+      Unix.set_close_on_exec masterFd;
+      Unix.set_close_on_exec slaveFd;
+      flush_all (); (* Clear buffers to avoid risk of double flushing by child.
+        Even this is not sufficient, strictly speaking, as there is a window
+        of opportunity to fill the buffer between flushing and calling fork. *)
       begin match Unix.fork () with
         0 ->
           begin try
+            (* Child process stderr must redirected as early as possible to
+               make sure all error output is captured and visible in GUI. *)
+            Unix.dup2 ~cloexec:false slaveFd Unix.stderr;
+            (* new_stderr will be used by parent process only. *)
+            if new_stderr <> Unix.stderr then safe_close new_stderr;
             Unix.close masterFd;
             ignore (Unix.setsid ());
             setControllingTerminal slaveFd;
@@ -309,14 +241,21 @@ let unix_create_session cmd args new_stdin new_stdout new_stderr =
             Unix.tcsetattr slaveFd Unix.TCSANOW tio;
             (* Redirect ssh authentication errors to controlling terminal,
                instead of new_stderr, so that they can be captured by GUI.
-               This will also redirect the remote stderr to GUI. *)
-            safe_close new_stderr;
+               This will inevitably also redirect the remote stderr to GUI
+               as ssh's own error output is mixed with remote stderr output. *)
             perform_redirections new_stdin new_stdout slaveFd;
             Unix.execvp cmd args (* never returns *)
           with Unix.Unix_error (e, s1, s2) ->
             Printf.eprintf "Error in create_session child: [%s] (%s) %s\n"
               s1 s2 (Unix.error_message e);
             flush stderr;
+            (* FIXME: this should be Unix._exit (available from OCaml 4.12)
+               which doesn't flush buffers (or run other exit handlers).
+               When [_exit] is eventually used then to _completely_ avoid risk
+               of double flushing, [Unix.write Unix.stderr] should be used
+               above instead of [eprintf]. Using [_exit] and not using any
+               [Stdlib.out_channel] will avoid all buffering and exit handler
+               issues. *)
             exit 127
           end
       | childPid ->
@@ -425,13 +364,13 @@ let handlePasswordRequests (fdIn, fdOut) callback isReady =
       let d = (Unix.gettimeofday ()) -. !time in
       if d < delay
         || Buffer.length scrollback = 0 then
-      (* HACK: Delay briefly (0.2 s, noticable to a human but not terrible
+      (* HACK: Delay briefly (0.2 s, noticeable to a human but not terrible
          either since some delay from the network is expected anyway) to allow
          time for output to be generated and read in as a whole. Otherwise, it
          may happen that 'Invalid password' and 'Please re-enter password'
          prompts are received separately, and this is not what we want.
          Loop by 0.01 s to let other threads run while not introducing
-         noticable latency to the user. *)
+         noticeable latency to the user. *)
       Lwt_unix.sleep (max (delay -. (max 0. d)) 0.01) >>= prompt
     else
       let query = extract () in
