@@ -104,13 +104,15 @@ let setColorPreference () =
   let envOk = try let _ = System.getenv "NO_COLOR" in false
     with Not_found -> true
   and termOk = try System.getenv "TERM" <> "dumb" with Not_found -> true
-  and ttyOk = (Unix.isatty Unix.stdin) && (Unix.isatty Unix.stderr) in
+  and ttyOk = (Unix.isatty Unix.stdout) && (Unix.isatty Unix.stderr) in
   let colorOk = envOk && termOk && ttyOk && not (Prefs.read dumbtty) in
   colorEnabled :=
     match Prefs.read colorMode with
     | `True    -> true
     | `False   -> false
-    | `Default -> colorOk && Sys.os_type <> "Win32"
+    | `Default -> colorOk && (not Sys.win32
+                    || (System.termVtCapable Unix.stdout
+                        && System.termVtCapable Unix.stderr))
 
 let color t =
   if not !colorEnabled then "" else
@@ -845,11 +847,58 @@ let doTransport reconItemList =
   in
   Uutil.setProgressPrinter showProgress;
 
-  Transport.logStart ();
+  let intrcount = ref 0 in
+  let sigtermHandler _ =
+    if !intrcount >= 3 then raise Sys.Break;
+    Abort.all ();
+    incr intrcount
+  in
+  let ctrlCHandler n =
+    sigtermHandler n;
+    if !intrcount = 1 then
+      let s = "\n\nUpdate propagation interrupted. It may take a while \
+        to stop.\nIf the process doesn't stop soon then wait or press \
+        Ctrl-C\n3 more times to force immediate termination.\n\n\n" in
+      (* Don't use [Printf.*printf] or [Format.*printf] (or other functions
+         which use [Stdlib.out_channel]) because this can cause a deadlock
+         with other outputting functions (in this case most likely at
+         [Util.set_infos] called in [showProgress]) before OCaml 4.12. *)
+      try Unix.write_substring Unix.stdout s 0 (String.length s) |> ignore
+      with Unix.Unix_error _ -> ()
+  in
+  let stopAtIntr f =
+    let signal_noerr signa behv =
+      try Some (Sys.signal signa behv)
+      with Sys_error _ | Invalid_argument _ -> None
+    in
+    let restore_noerr signa = function
+    | Some prevSig -> ignore (signal_noerr signa prevSig)
+    | None -> ()
+    in
+    let prevSigInt = signal_noerr Sys.sigint (Signal_handle ctrlCHandler) in
+    let prevSigTerm = signal_noerr Sys.sigterm (Signal_handle sigtermHandler) in
+    let restoreSig () =
+      (* Set handlers will still raise [Sys.Break]; can ignore errors here. *)
+      restore_noerr Sys.sigint prevSigInt;
+      restore_noerr Sys.sigterm prevSigTerm
+    in
+
+    try f (); restoreSig ()
+    with e ->
+      let origbt = Printexc.get_raw_backtrace () in
+      restoreSig ();
+      Printexc.raise_with_backtrace e origbt
+  in
+
+  Uicommon.transportStart ();
   let fFailedPaths = ref [] in
   let fPartialPaths = ref [] in
-  let uiWrapper i item f =
-    Lwt.try_bind f
+  let notstarted = ref (Array.length items) in
+  let uiWrapper i item =
+    Lwt.try_bind
+      (fun () -> decr notstarted;
+                 Transport.transportItem item.ri
+                   (Uutil.File.ofLine i) verifyMerge)
       (fun () ->
          if partiallyProblematic item.ri && not (problematic item.ri) then
            fPartialPaths := item.ri.path1 :: !fPartialPaths;
@@ -870,35 +919,16 @@ let doTransport reconItemList =
             return ()
         | _ ->
             fail e) in
-  let im = Array.length items in
-  let rec loop i actions pRiThisRound =
-    if i < im then begin
-      let item = items.(i) in
-      let actions =
-        if pRiThisRound item.ri then
-          uiWrapper i item
-            (fun () -> Transport.transportItem item.ri
-                         (Uutil.File.ofLine i) verifyMerge)
-          :: actions
-        else
-          actions
-      in
-      loop (i + 1) actions pRiThisRound
-    end else
-      actions
-  in
-  Lwt_unix.run
-    (let actions = loop 0 [] (fun ri -> not (Common.isDeletion ri)) in
-     Lwt_util.join actions);
-  Lwt_unix.run
-    (let actions = loop 0 [] Common.isDeletion in
-     Lwt_util.join actions);
-  Transport.logFinish ();
+  stopAtIntr begin fun () ->
+    Uicommon.transportItems items (fun {ri; _} -> not (Common.isDeletion ri)) uiWrapper;
+    Uicommon.transportItems items (fun {ri; _} -> Common.isDeletion ri) uiWrapper
+  end;
+  Uicommon.transportFinish ();
 
   Uutil.setProgressPrinter (fun _ _ _ -> ());
   Util.set_infos "";
 
-  (Safelist.rev !fFailedPaths, Safelist.rev !fPartialPaths)
+  (Safelist.rev !fFailedPaths, Safelist.rev !fPartialPaths, !notstarted, !intrcount > 0)
 
 let setWarnPrinterForInitialization()=
   Util.warnPrinter :=
@@ -936,8 +966,8 @@ let formatStatus major minor =
     s
 
 let rec interactAndPropagateChanges prevItemList reconItemList
-            : bool * bool * bool * (Path.t list)
-              (* anySkipped?, anyPartial?, anyFailures?, failingPaths *) =
+      : bool * bool * bool * bool * (Path.t list)
+        (* anySkipped?, anyPartial?, anyFailures?, anyCancels?, failingPaths *) =
   let (proceed,newReconItemList) = interact prevItemList reconItemList in
   let (updatesToDo, skipped) =
     Safelist.fold_left
@@ -945,21 +975,41 @@ let rec interactAndPropagateChanges prevItemList reconItemList
         if problematic ri then (howmany, skipped + 1)
         else (howmany + 1, skipped))
       (0, 0) newReconItemList in
+  let doTransp newReconItemList =
+    try
+      doTransport newReconItemList
+    with e ->
+      let origbt = Printexc.get_raw_backtrace () in
+      let summary =
+        "\nSynchronization "
+          ^ (color `Failure)
+          ^ (match e with Sys.Break -> "interrupted" | _ -> "failed")
+          ^ (color `Reset)
+          ^ (try let tm = Util.localtime (Util.time ()) in
+             Printf.sprintf " at %02d:%02d:%02d"
+               tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec with _ -> "")
+          ^ (match e with Sys.Break -> " by user request" | _ -> " due to a fatal error")
+          ^ "\n\n"
+      in
+      Trace.log_color summary;
+      Printexc.raise_with_backtrace e origbt
+  in
   let doit() =
     if not (Prefs.read Globals.batch || Prefs.read Trace.terse) then newLine();
     if not (Prefs.read Trace.terse) then Trace.status "Propagating updates";
     let timer = Trace.startTimer "Transmitting all files" in
-    let (failedPaths, partialPaths) = doTransport newReconItemList in
+    let (failedPaths, partialPaths, notstarted, intr) = doTransp newReconItemList in
     let failures = Safelist.length failedPaths in
     let partials = Safelist.length partialPaths in
     Trace.showTimer timer;
     if not (Prefs.read Trace.terse) then Trace.status "Saving synchronizer state";
     Update.commitUpdates ();
-    let trans = updatesToDo - failures in
+    let trans = updatesToDo - notstarted - failures in
     let summary =
       Printf.sprintf
-       "Synchronization %s at %s  (%d item%s transferred, %s%s, %s)"
-       (if failures = 0 then (color `Success) ^ "complete" ^ (color `Reset) else (color `Failure) ^ "incomplete" ^ (color `Reset))
+       "Synchronization %s at %s  (%d item%s transferred, %s%s, %s%s)"
+       (if failures = 0 && notstarted = 0 then (color `Success) ^ "complete" ^ (color `Reset)
+        else (color `Failure) ^ "incomplete" ^ (color `Reset))
        (let tm = Util.localtime (Util.time()) in
         Printf.sprintf "%02d:%02d:%02d"
           tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec)
@@ -969,7 +1019,8 @@ let rec interactAndPropagateChanges prevItemList reconItemList
         else
           "")
        (if skipped = 0 then "0 skipped" else (color `Information) ^ (Printf.sprintf "%d skipped" skipped) ^ (color `Reset))
-       (if failures = 0 then "0 failed" else (color `Failure) ^ (Printf.sprintf "%d failed" failures) ^ (color `Reset)) in
+       (if failures = 0 then "0 failed" else (color `Failure) ^ (Printf.sprintf "%d failed" failures) ^ (color `Reset))
+       (if notstarted = 0 then "" else ", " ^ (color `Information) ^ (Printf.sprintf "%d not started" notstarted) ^ (color `Reset)) in
     Trace.log_color (summary ^ "\n");
     if skipped>0 then
       Safelist.iter
@@ -990,7 +1041,8 @@ let rec interactAndPropagateChanges prevItemList reconItemList
       Safelist.iter
         (fun p -> alwaysDisplayAndLog ("  failed: " ^ (Path.toString p)))
         failedPaths;
-    (skipped > 0, partials > 0, failures > 0, failedPaths) in
+    if intr then raise Sys.Break; (* Make sure repeat mode is stopped *)
+    (skipped > 0, partials > 0, failures > 0, notstarted > 0, failedPaths) in
   if updatesToDo = 0 then begin
     (* BCP (3/09): We need to commit the archives even if there are
        no updates to propagate because some files (in fact, if we've
@@ -999,7 +1051,7 @@ let rec interactAndPropagateChanges prevItemList reconItemList
     (* JV (5/09): Don't save the archive in repeat mode as it has some
        costs and its unlikely there is much change to the archives in
        this mode. *)
-    if !Update.foundArchives && Prefs.read Uicommon.repeat = "" then
+    if !Update.foundArchives && Prefs.read Uicommon.repeat = `NoRepeat then
       Update.commitUpdates ();
     display "No updates to propagate\n";
     if skipped > 0 then begin
@@ -1025,7 +1077,7 @@ let rec interactAndPropagateChanges prevItemList reconItemList
          | _ -> ())
         newReconItemList
       end;
-    (skipped > 0, false, false, [])
+    (skipped > 0, false, false, false, [])
   end else if proceed=ProceedImmediately then begin
     doit()
   end else
@@ -1107,6 +1159,10 @@ let checkForDangerousPath dangerousPaths =
     end
   end
 
+let displayWaitMessage () =
+  if not (Prefs.read silent) then
+    Util.msg "%s\n" (Uicommon.contactingServerMsg ())
+
 let synchronizeOnce ?wantWatcher pathsOpt =
   let showStatus path =
     if path = "" then Util.set_infos "" else
@@ -1121,10 +1177,7 @@ let synchronizeOnce ?wantWatcher pathsOpt =
     let c = "-\\|/".[truncate (mod_float (4. *. Unix.gettimeofday ()) 4.)] in
     Util.set_infos (Format.sprintf "%c %s" c path)
   in
-  Uicommon.refreshConnection
-    ~displayWaitMessage:(fun () -> if not (Prefs.read silent)
-                         then Util.msg "%s\n" (Uicommon.contactingServerMsg()))
-    ~termInteract:None;
+  Uicommon.connectRoots ~displayWaitMessage ();
   Trace.status "Looking for changes";
   if not (Prefs.read Trace.terse) && (Prefs.read Trace.debugmods = []) then
     Uutil.setUpdateStatusPrinter (Some showStatus);
@@ -1142,7 +1195,7 @@ let synchronizeOnce ?wantWatcher pathsOpt =
 
   if not !Update.foundArchives then Update.commitUpdates ();
   if reconItemList = [] then begin
-    if !Update.foundArchives && Prefs.read Uicommon.repeat = "" then
+    if !Update.foundArchives && Prefs.read Uicommon.repeat = `NoRepeat then
       Update.commitUpdates ();
     (if anyEqualUpdates then
       Trace.status ("Nothing to do: replicas have been changed only "
@@ -1152,12 +1205,85 @@ let synchronizeOnce ?wantWatcher pathsOpt =
     (Uicommon.perfectExit, [])
   end else begin
     checkForDangerousPath dangerousPaths;
-    let (anySkipped, anyPartial, anyFailures, failedPaths) =
+    let (anySkipped, anyPartial, anyFailures, anyCancel, failedPaths) =
       interactAndPropagateChanges [] reconItemList in
-    let exitStatus = Uicommon.exitCode(anySkipped || anyPartial,anyFailures) in
+    let exitStatus = Uicommon.exitCode (anySkipped || anyPartial || anyCancel, anyFailures) in
     (exitStatus, failedPaths)
   end
 
+(* ------------ Safe termination between synchronizations ------------ *)
+
+let safeStopReqd, requestSafeStop =
+  let safeStopReqd = ref false in
+  (* [safeStopReqd] can only go from false to true;
+     it must never be changed from true to false. *)
+  let isRequested () = !safeStopReqd
+  and request () = safeStopReqd := true in
+  isRequested, request
+
+(*** Requesting safe termination by signals ***)
+
+let set_signal_noerr signa nm behv =
+  try Sys.set_signal signa behv; true
+  with Invalid_argument _ | Sys_error _ as e ->
+    Trace.logonly
+      ("Warning: " ^ nm ^ " handler not set: " ^ (Printexc.to_string e) ^ "\n");
+    false
+
+let stopPipe = ref None
+
+let setupSafeStop () =
+  if supportSignals then begin
+    let safeStop _ =
+      if not (safeStopReqd ()) then begin
+        requestSafeStop ();
+        (* Interrupt the interruptible sleep *)
+        match !stopPipe with
+        | Some (i, o) -> Unix.close o; Lwt_unix.close i
+        | None -> ()
+      end
+    in
+    Util.blockSignals [Sys.sigusr2] (fun () ->
+    let ok = set_signal_noerr Sys.sigusr2 "SIGUSR2" (Signal_handle safeStop) in
+    if ok then stopPipe := Some (Lwt_unix.pipe_in ~cloexec:true ()))
+  end
+
+let safeStopRequested () =
+  safeStopReqd ()
+
+(*** Sleep interruptible by a termination request ***)
+
+let safeStopWait =
+  let safeStopWait_aux () =
+    let readStop =
+      match !stopPipe with
+      | None -> Lwt.wait ()
+      | Some (i, _) -> Lwt_unix.wait_read i
+    in
+    let readFail = function
+      | Unix.Unix_error (EBADF, _, _) -> Lwt.return (requestSafeStop ())
+      | e -> Lwt.fail e
+    in
+    let rec loop () =
+      Lwt.catch
+        (fun () -> readStop) readFail >>= fun () ->
+      if not (safeStopRequested ()) then
+        Lwt_unix.sleep 0.15 >>= loop
+      else
+        Lwt.return ()
+    in
+    loop ()
+  in
+  let wt = ref None in
+  fun () ->
+    match !wt with
+    | Some t -> t
+    | None -> let t = safeStopWait_aux () in wt := Some t; t
+
+let interruptibleSleepf dt =
+  Lwt_unix.run (Lwt.choose [Lwt_unix.sleep dt; safeStopWait ()])
+let interruptibleSleep dt = interruptibleSleepf (float dt)
+
 (* ----------------- Filesystem watching mode ---------------- *)
 
 let watchinterval = 1.    (* Minimal interval between two synchronizations *)
@@ -1178,12 +1304,17 @@ let waitForChanges t =
     Lwt_unix.run
       (Globals.allRootsMap (fun r -> Lwt.return (waitForChangesRoot r ()))
          >>= fun l ->
-       Lwt.choose (timeout @ l))
+       Lwt.choose (timeout @ l @ [safeStopWait ()]))
   end
 
-let synchronizePathsFromFilesystemWatcher () =
-  let rec loop isStart delayInfo =
+let synchronizePathsFromFilesystemWatcher fullintv =
+  let fullinterval = match fullintv with None -> 1e20 | Some i -> float i in
+  let rec loop lastFull delayInfo =
     let t = Unix.gettimeofday () in
+    let sinceFull = t -. lastFull in
+    let isFull = sinceFull > fullinterval in
+    let lastFull = if isFull then t else lastFull in
+    let nextFull = lastFull +. fullinterval in
     let (delayedPaths, readyPaths) =
       PathMap.fold
         (fun p (t', _) (delayed, ready) ->
@@ -1192,7 +1323,7 @@ let synchronizePathsFromFilesystemWatcher () =
     in
     let (exitStatus, failedPaths) =
       synchronizeOnce ~wantWatcher:true
-        (if isStart then None else Some (readyPaths, delayedPaths))
+        (if isFull then None else Some (readyPaths, delayedPaths))
     in
     (* After a failure, we retry at once, then use an exponential backoff *)
     let delayInfo =
@@ -1210,13 +1341,13 @@ let synchronizePathsFromFilesystemWatcher () =
         PathMap.empty
         (Safelist.append delayedPaths failedPaths)
     in
-    Lwt_unix.run (Lwt_unix.sleep watchinterval);
+    interruptibleSleepf watchinterval;
     let nextTime =
-      PathMap.fold (fun _ (t, d) t' -> min t t') delayInfo 1e20 in
-    waitForChanges nextTime;
-    loop false delayInfo
+      PathMap.fold (fun _ (t, d) t' -> min t t') delayInfo nextFull in
+    if not (safeStopRequested ()) then waitForChanges nextTime;
+    if safeStopRequested () then exitStatus else loop lastFull delayInfo
   in
-  loop true PathMap.empty
+  loop 0. PathMap.empty
 
 (* ----------------- Repetition ---------------- *)
 
@@ -1225,37 +1356,34 @@ let synchronizeUntilNoFailures repeatMode =
   let rec loop triesLeft pathsOpt =
     let (exitStatus, failedPaths) =
       synchronizeOnce ~wantWatcher pathsOpt in
-    if failedPaths <> [] && triesLeft <> 0 then begin
+    if failedPaths <> [] && triesLeft <> 0
+         && not (repeatMode && safeStopRequested ()) then begin
       loop (triesLeft - 1) (Some (failedPaths, []))
     end else begin
       exitStatus
     end in
   loop (Prefs.read Uicommon.retry) None
 
-let rec synchronizeUntilDone () =
-  let repeatinterval =
-    if Prefs.read Uicommon.repeat = "" then -1 else
-    try int_of_string (Prefs.read Uicommon.repeat)
-    with Failure _ ->
-      (* If the 'repeat' pref is not a valid number, switch modes... *)
-      if Prefs.read Uicommon.repeat = "watch" then
-        synchronizePathsFromFilesystemWatcher()
-      else
-        raise (Util.Fatal ("Value of 'repeat' preference ("
-                           ^Prefs.read Uicommon.repeat
-                           ^") should be either a number or 'watch'\n")) in
-
+let rec synchronizeUntilDone repeatinterval =
   let exitStatus = synchronizeUntilNoFailures(repeatinterval >= 0) in
-  if repeatinterval < 0 then
+  if repeatinterval < 0 || safeStopRequested () then
     exitStatus
   else begin
     (* Do it again *)
     Trace.status (Printf.sprintf
        "\nSleeping for %d seconds...\n" repeatinterval);
-    Unix.sleep repeatinterval;
-    synchronizeUntilDone ()
+    interruptibleSleep repeatinterval;
+    if safeStopRequested () then exitStatus else synchronizeUntilDone repeatinterval
   end
 
+let synchronizeUntilDone () =
+  match Prefs.read Uicommon.repeat with
+  | `Watch -> synchronizePathsFromFilesystemWatcher None
+  | `WatchAndInterval i -> synchronizePathsFromFilesystemWatcher (Some i)
+  | `Interval i -> synchronizeUntilDone i
+  | `NoRepeat -> synchronizeUntilDone (-1)
+  | `Invalid (_, e) -> raise e
+
 (* ----------------- Startup ---------------- *)
 
 let profmgrPrefName = "i"
@@ -1393,8 +1521,10 @@ let rec start interface =
   if interface <> Uicommon.Text then
     Util.msg "This Unison binary only provides the text GUI...\n";
   begin try
+    Sys.catch_break true;
     (* Just to make sure something is there... *)
     setWarnPrinterForInitialization();
+    setupSafeStop ();
     let errorOut s =
       Util.msg "%s%s%s\n" Uicommon.shortUsageMsg profmgrUsageMsg s;
       exit 1
@@ -1411,18 +1541,20 @@ let rec start interface =
           end
       | Ok (Some s) -> s
     in
-    Uicommon.initPrefs
-      ~profileName
-      ~displayWaitMessage:
-      (fun () -> setWarnPrinter();
-                 if Prefs.read silent then Prefs.set Trace.terse true;
-                 if not (Prefs.read silent)
-                 then Util.msg "%s\n" (Uicommon.contactingServerMsg()))
-      ~promptForRoots:
-      (fun () -> errorOut "")
-      ~termInteract:
-      None
-      ();
+    Uicommon.initPrefs ~profileName ~promptForRoots:(fun () -> errorOut "") ()
+  with e ->
+    handleException e;
+    exit Uicommon.fatalExit
+  end;
+
+  (* Uncaught exceptions up to this point are non-recoverable, treated
+     as permanent and will inevitably exit the process. Uncaught exceptions
+     from here onwards are treated as potentially temporary or recoverable.
+     The process does not have to exit if in repeat mode and can try again. *)
+  begin try
+    if Prefs.read silent then Prefs.set Trace.terse true;
+
+    Uicommon.connectRoots ~displayWaitMessage ();
 
     if Prefs.read Uicommon.testServer then exit 0;
 
@@ -1439,7 +1571,7 @@ let rec start interface =
       Prefs.set dumbtty true;
       Trace.sendLogMsgsToStderr := false;
     end;
-    if Prefs.read Uicommon.repeat <> "" then begin
+    if Prefs.read Uicommon.repeat <> `NoRepeat then begin
       Prefs.set Globals.batch true;
     end;
     setColorPreference ();
@@ -1464,23 +1596,38 @@ let rec start interface =
       handleException Sys.Break;
       exit Uicommon.fatalExit
     end
+  | e when breakRepeat e -> begin
+      handleException e;
+      exit Uicommon.fatalExit
+    end
   | e -> begin
       (* If any other bad thing happened and the -repeat preference is
          set, then restart *)
-      (* JV: it seems safer to just abort here, as we don't know in which
-         state Unison is; for instance, if the connection is lost, there
-         is no point in restarting as Unison will currently not attempt to
-         establish a new connection. *)
       handleException e;
-      if false (*Prefs.read Uicommon.repeat <> ""*) then begin
-        Util.msg "Restarting in 10 seconds...\n";
-        Unix.sleep 10;
-        start interface
-      end else
-        exit Uicommon.fatalExit
+      if Prefs.read Uicommon.repeat = `NoRepeat
+          || Prefs.read Uicommon.runtests then
+        exit Uicommon.fatalExit;
+
+      Util.msg "\nRestarting in 10 seconds...\n\n";
+      begin try interruptibleSleep 10 with Sys.Break -> exit Uicommon.fatalExit end;
+      if safeStopRequested () then exit Uicommon.fatalExit else start interface
     end
   end
 
+(* Though in some cases we could, there's no point in recovering
+   and continuing at any of these exceptions. *)
+and breakRepeat = function
+  (* Programming errors *)
+  | Assert_failure _
+  | Match_failure _
+  | Invalid_argument _
+  | Fun.Finally_raised _
+  (* Async exceptions *)
+  | Out_of_memory
+  | Stack_overflow
+  | Sys.Break -> true
+  | _ -> false
+
 let defaultUi = Uicommon.Text
 
 end
